from typing import List, Optional
import torch


def collate_tokens(
        values: List[torch.Tensor], pad_idx: int, left_pad: bool = False,
        pad_to_length: Optional[int] = None, pad_to_multiple: int = 1
):
    """Convert a list of 1d tensors into a padded 2d tensor."""
    size = max(v.size(0) for v in values)
    size = size if pad_to_length is None else max(size, pad_to_length)

    if pad_to_multiple != 1 and size % pad_to_multiple != 0:
        size = int(((size-0.1)//pad_to_multiple + 1) * pad_to_multiple)

    res = values[0].new_full((len(values), size), fill_value=pad_idx)

    def copy_tensor(src, dst):
        assert dst.numel() == src.numel()
        dst.copy_(src)

    for i, v in enumerate(values):
        copy_tensor(v, res[i][size - len(v):] if left_pad else res[i][:len(v)])
    return res


def collate_descendants(
        descendants: List[torch.Tensor], left_pad: bool = False, pad_idx: int = 0,
        pad_to_length: Optional[int] = None, pad_to_multiple: int = 1
):
    """Convert a list of 1d tensors of descendants into a padded 2d tensor. It will
    increment all values by (pad_idx + 1).
    """
    return collate_tokens(
        [v + (pad_idx + 1) for v in descendants],
        pad_idx, left_pad, pad_to_length, pad_to_multiple
    )


def collate_parents(
        parents: List[torch.Tensor], left_pad: bool = False, pad_idx: int = -2,
        pad_to_length: Optional[int] = None, pad_to_multiple: int = 1
):
    """Convert a list of 1d tensors of parent indices into a padded 2d tensor.
        Increments the tokens by the amount of added padding tokens.

        If pad_idx is part of the tensor, it will not be incremented!

        [ [-1, 0, 1, 0],
          [-1, 0, 1] ]
        ->
        [ [-1,  0, 1, 0],
          [-1, -1, 1, 2] ] if pad_idx = -1 and left_pad

     """
    size = max(v.size(0) for v in parents)
    size = size if pad_to_length is None else max(size, pad_to_length)

    if pad_to_multiple != 1 and size % pad_to_multiple != 0:
        size = int(((size - 0.1) // pad_to_multiple + 1) * pad_to_multiple)

    res = parents[0].new_full((len(parents), size), fill_value=pad_idx)

    def copy_tensor(src, dst):
        assert dst.numel() == src.numel()
        dst.copy_(src)

    for i, v in enumerate(parents):
        num_pad = size - len(v)
        if num_pad > 0:
            if left_pad:  # parent indices change if we left pad
                mask = (v != pad_idx)  # * (v != eos_idx)
                v[mask] += num_pad

        copy_tensor(v, res[i][num_pad:] if left_pad else res[i][:len(v)])

    return res


def collate_tree(
        batch, left_pad: bool = False, pad_idx: int = -1,
        pad_to_length: Optional[int] = None, pad_to_multiple: int = 1
):

    parents = collate_parents(
        [s["tree"].pop("parents") for s in batch],
        left_pad, pad_to_length=pad_to_length, pad_to_multiple=pad_to_multiple
    )
    descendants = collate_descendants(
        [s["tree"].pop("descendants") for s in batch],
        left_pad, pad_to_length=pad_to_length, pad_to_multiple=pad_to_multiple
    )
    tokens = collate_tokens(
        [s["tree"].pop("tokens") for s in batch],
        pad_idx, left_pad, pad_to_length, pad_to_multiple
    )
    labels = collate_descendants(
        [s["labels"] for s in batch],
        left_pad, pad_idx=-1, pad_to_length=pad_to_length, pad_to_multiple=pad_to_multiple
    )

    return {
        "labels": labels,  # process remaining items
        "tree": {
            "tokens": tokens,
            "descendants": descendants,
            "parents": parents,
        },
    }





from typing import Optional, Generator, Union, Tuple

import torch

import tensortree


def mask_layer(node_incidences: torch.BoolTensor) -> Generator[torch.BoolTensor, None, None]:
    """
    Generates masks, which traverse each layer of the tree
    from bottom to top.
        1. Mask over all leaves.
        2. Mask over all parents of the leaves.
        3. .. parents of parents ...

    :param node_incidence_matrix:
    :return:
    """
    batched = node_incidences.ndim == 3
    if not batched:
        node_incidences = node_incidences[None, :, :]

    node_indicences = node_incidences.clone()

    while node_indicences.any():
        leaves = node_indicences.sum(-2) == 1
        yield leaves if batched else leaves.squeeze()
        node_indicences.masked_fill_(leaves[:, :, None], 0)


def mask_level(
    node_incidences, reverse=False, return_level: bool = False
) -> Generator[Union[torch.BoolTensor, Tuple[int, torch.BoolTensor]], None, None]:
    """
    Generates masks, which traverse each level of the tree
    from bottom to top.
        1. Mask over tokens with lowest level.
        ...

    :param node_incidence_matrix:
    :return:
    """
    all_levels = tensortree.levels(node_incidences)
    max_level = all_levels.max()
    for level in range(1, max_level + 1):
        level = level if reverse else (max_level + 1) - level
        mask = all_levels == level

        yield (level, mask,) if return_level else mask



from itertools import takewhile, repeat
import pathlib as pl
from typing import Union, List, Optional, Sequence, Any

import numpy as np
import torch

import functools


def validate_index(_func=None, allow_none: bool = False):
    """ Should be only used inside TensorTree and on functions that receive
    a node_idx as the first argument after self.
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, node_idx, *args, **kwargs):

            if node_idx is None:
                if not allow_none:
                    raise IndexError(f"Index {node_idx} is not allowed in this method.")
            else:
                if node_idx < 0 or node_idx >= len(self):
                    raise IndexError(
                        f"Index {node_idx} is out of bounds for this"
                        f" {'sub' if self.is_subtree() else ''}tree with {len(self)} nodes."
                    )

            return func(self, node_idx, *args, **kwargs)
        return wrapper

    if _func is None:
        return decorator  # 2
    else:
        return decorator(_func)


def linecount(filename: Union[pl.Path, str]) -> int:
    with open(filename, 'rb') as f:
        bufgen = takewhile(
            lambda x: x,
            (
                f.raw.read(1024 * 1024) for _ in repeat(None)
            )
        )
        return sum(
            buf.count(b'\n') for buf in bufgen
        )


def get_project_root():
    return pl.Path(__file__).parent.parent


def to_torch(some_sequence: Sequence[Any]) -> torch.Tensor:
    if isinstance(some_sequence, torch.Tensor):
        return some_sequence
    elif isinstance(some_sequence, np.ndarray):
        return torch.from_numpy(some_sequence)
    else:
        return torch.tensor(some_sequence)  # may raise additional errors


def is_tensor_type(sequence: Union[torch.Tensor, np.ndarray, Any]) -> bool:
    return isinstance(sequence, (torch.Tensor, np.ndarray))


def to_matmul_compatibility(x: torch.Tensor) -> torch.Tensor:
    """ Brings a tensor into a matmul compatible dtype """
    if x.is_cuda:
        # bmm doesnt work with integers
        return x.float()
    elif x.dtype == torch.bool:
        # mm doesnt work on boolean arrays
        return x.long()
    elif not x.is_floating_point and (x.size(-1) >= torch.iinfo(x.dtype).max):
        # prevent overflow if we multiply uint8
        return x.long()

    return x


def apply_pad_mask_to_higher_dim_tensor(x: torch.Tensor, pad_idx: int, pad_mask: torch.Tensor):
    assert x.ndimension() == (pad_mask.ndimension() + 1), "x should have one dimension more than pad_mask"

    if pad_mask.ndimension() == 2:
        x[pad_mask] = pad_idx
        x[pad_mask[:, None, :].expand_as(x)] = pad_idx
    elif pad_mask.ndimension() == 1:
        # then x is 2-dimensional
        x[pad_mask] = pad_idx
        x[:, pad_mask] = pad_idx
    else:
        raise ValueError("pad mask should be either 1d or 2d")

    return x


def print_embed_overlap(embed_dict, vocab_dict):
    embed_keys = set(embed_dict.keys())
    vocab_keys = set(vocab_dict.symbols)
    overlap = len(embed_keys & vocab_keys)
    print("found {}/{} types in embedding file".format(overlap, len(vocab_dict)))


def parse_embedding(embed_path):
    """Parse embedding text file into a dictionary of word and embedding tensors.

    The first line can have vocabulary size and dimension. The following lines
    should contain word and embedding separated by spaces.

    Example:
        2 5
        the -0.0230 -0.0264  0.0287  0.0171  0.1403
        at -0.0395 -0.1286  0.0275  0.0254 -0.0932
    """
    print("Loading embedding file from", embed_path)
    from tqdm import tqdm
    embed_dict = {}
    with open(embed_path) as f_embed:
        next(f_embed)  # skip header
        for line in tqdm(f_embed):
            pieces = line.rstrip().split(" ")
            embed_dict[pieces[0]] = torch.Tensor(
                [float(weight) for weight in pieces[1:]]
            )
    return embed_dict


def load_embedding(embed_dict, vocab, embedding):
    for idx in range(len(vocab)):
        token = vocab[idx]
        if token in embed_dict:
            embedding.weight.data[idx] = embed_dict[token]
    return embedding


from dataclasses import dataclass
from typing import Any, Callable, Generator, Optional, Union

from torch import Tensor

import tensortree


@dataclass
class Row:
    pre: str
    fill: str
    node: str


@dataclass
class Style:
    vertical: str
    cont: str
    end: str


@dataclass
class AsciiStyle(Style):
    vertical: str = '|   '
    cont: str = '|-- '
    end: str = '+-- '


@dataclass
class ContStyle(Style):
    vertical: str = '\u2502   '
    cont: str = '\u251c\u2500\u2500 '
    end: str = '\u2514\u2500\u2500 '


@dataclass
class ContRoundStyle(Style):
    vertical: str = '\u2502   '
    cont: str = '\u251c\u2500\u2500 '
    end: str = '\u2570\u2500\u2500 '


@dataclass
class DoubleStyle(Style):
    vertical: str = '\u2551   '
    cont: str = '\u2560\u2550\u2550 '
    end: str = '\u255a\u2550\u2550 '


def format_tree(
        tree, max_nodes: Optional[int] = None, node_renderer: Callable[[Any], str] = str,
        style: Union[Style] = ContRoundStyle,
) -> str:
    """
    Pretty prints a tree up to `max_nodes`. Define a node_renderer for custom node types (e.g. Dictionaries).
    :param max_nodes: Render up to this amount of nodes.
    :param node_renderer: A function that outputs a string.
    :param style: Style the tree.
    :return:
    """
    def format_nodes() -> Generator[Row, None, None]:
        incidences = tree.node_incidence_matrix()
        levels = tensortree.levels(incidences).tolist()
        active_levels = set()

        def make_row(token, node_idx, level, active_levels, fill, replace_token: bool = False):
            pre = "".join(style.vertical if l in active_levels else "    " for l in range(level - 1))

            if level == 0:
                fill = ""

            if isinstance(token, Tensor):
                token = token.item()

            if replace_token:
                token = "[...]"
            else:
                token = node_renderer(token)

            return Row(pre, fill, f"{node_idx}. {token}")

        for node_idx, (token, lev) in enumerate(zip(tree.node_data, levels), start=0):
            fill = style.end

            if tree.has_sibling(node_idx, check_left=False, check_right=True):
                fill = style.cont
                if not tree.is_leaf(node_idx):
                    active_levels.add((lev - 1))
            elif (lev - 1) in active_levels:
                active_levels.remove(lev - 1)

            row = make_row(token, node_idx, lev, active_levels, fill)

            yield row

            # code for stopping at max_nodes and finish open levels below
            # maybe we break before printing this row, to print a shorter version of this tree
            if max_nodes is not None and node_idx >= (max_nodes - len(active_levels) - 2):

                # this was the last node of this branch
                if fill == style.end:
                    next_node = tree.next_node_not_in_branch(node_idx)
                elif fill == style.cont:
                    next_node = tree.right_sibling(node_idx)
                active_levels.add(None)

                while active_levels and next_node:
                    yield make_row(None, next_node, levels[next_node], active_levels, style.end, replace_token=True)
                    next_node = tree.step_out(next_node)
                    some_lev = active_levels.pop()

                break

    return "\n".join(
        f"{row.pre}{row.fill}{row.node}" for row in format_nodes()
    )



from tensortree.tree import TensorTree, TreeStorage, LabelType, TensorType, tree
from tensortree.operations import *
from tensortree.iterators import *
from tensortree.collate import collate_tokens, collate_parents, collate_descendants



from typing import Union

from nltk.tree import Tree as NLTKTree

from tensortree import TensorTree

def from_nltk(tree: Union[NLTKTree, str]) -> TensorTree:
    """

    :param tree: a NLTK Tree object or a string
    :param parent: a parent astlib Tree
    :return:
    """
    parents = []

    def traverse(tree):
        if isinstance(tree, NLTKTree):
            root = cls(
                name=tree.label(),
                _is_terminal=False,
                parent=parent
            )
            for node in tree:
                cls.from_nltk(node, parent=root)

        else:
            cls(
                name=tree,
                _is_terminal=True,
                parent=parent
            )

    return root


from spacy.tokens.doc import Doc


def from_spacy(sentence: Doc) -> TensorTree:
    """

    :param tree: a NLTK Tree object or a string
    :param parent: a parent astlib Tree
    :return:
    """
    parents = []
    node_data = []

    def traverse(token, parent=-1):
        print(token.text)
        data = {
            "token": token.text,
            "position": token.head.pos_,
            "relation": token.dep_,

        }
        node_data.append(data)
        parents.append(parent)

        node_idx = len(parents) - 1

        for child in token.children:
            traverse(child, parent=node_idx)

    traverse(sentence.root)
    print(parents)

    return TensorTree.from_array(node_data=node_data, parents=parents)


from typing import Union, Any, Optional, Sequence

import numpy as np
import torch

import tensortree
from tensortree import TensorTree
from tensortree.utils import is_tensor_type, to_matmul_compatibility, apply_pad_mask_to_higher_dim_tensor


def node_incidence_matrix(
        descendants: torch.Tensor, pad_idx: int = -1, pad_mask: Optional[torch.Tensor] = None
) -> torch.BoolTensor:
    """
    Computes the node incidence matrix between nodes based on the descendants array.
    This array may be batched, then it is created for all trees at once.

     1D or 2D tensor of descendants [seqlen] or [bsz x seqlen].

    If the tensor has been padded, this function assumes descendants
     has been incremented by `pad_idx + 1` (to make sure pad_idx
     does not appear in the sequence).

    Example:
    >>> node_incidence_matrix(
    >>>    descendants=torch.tensor(
    >>>        [[6, 2, 4, 2, 2],
    >>>         [1, 5, 2, 3, 2],
    >>>         [1, 5, 2, 3, 2]]
    >>>    ),
    >>>    pad_idx=1
    >>> )

    OUTPUT:
    >>> tensor([[[ True, False, False, False, False],
    >>>          [ True,  True, False, False, False],
    >>>          [ True, False,  True, False, False],
    >>>          [ True, False,  True,  True, False],
    >>>          [ True, False,  True, False,  True]],
    >>>         [[False, False, False, False, False],
    >>>          [False,  True, False, False, False],
    >>>          [False,  True,  True, False, False],
    >>>          [False,  True, False,  True, False],
    >>>          [False,  True, False,  True,  True]],
    >>>         [[False, False, False, False, False],
    >>>          [False,  True, False, False, False],
    >>>          [False,  True,  True, False, False],
    >>>          [False,  True, False,  True, False],
    >>>          [False,  True, False,  True,  True]]])

    Every row contains the nodes that lay on the path to root
     for node at row. True if a node lays on the path to root, False otherwise.

    """
    assert descendants.ndimension() <= 2, "Wrong dimensions on descendants tensor"
    batched = descendants.ndimension() == 2
    B = descendants.size(0) if batched else 1
    S = descendants.size(-1)

    # make it work also with 1D tensors
    if not batched:
        descendants = descendants[None, :]

        if pad_mask is not None:
            pad_mask = pad_mask[None, :]  # unsqueeze also pad mask

    descendants = descendants.clone()

    # compute padding_masks before calculating the correct descendants, afterwards we can't distinguish
    pad_mask = (descendants == pad_idx) if pad_mask is None else pad_mask
    token_mask = ~pad_mask

    positions = torch.arange(end=S, device=descendants.device)
    reverse_range = torch.flip(positions, dims=(0,))

    positions = positions.expand_as(descendants)
    reverse_range = reverse_range.expand_as(descendants)

    # every node is on its own path to root
    descendants[token_mask] = descendants[token_mask] - (pad_idx + 1) + 1
    descendants[pad_mask] = reverse_range[pad_mask]

    reps = descendants.new_zeros((B, 3*S,))
    # num repetitions of pad before diag
    reps[:, ::3] = positions
    # num repetitions of position after diag
    reps[:, 1::3] = descendants
    # num repetitions of pad after diag
    reps[:, 2::3] = (descendants - S).abs() - positions

    val_mask = pad_mask.new_full(reps.shape, False)
    val_mask[:, 1::3] = token_mask
    mat = torch.repeat_interleave(
        val_mask.view(-1),
        reps.view(-1)
    ).view(B, S, S).transpose(1, 2)

    return mat if batched else mat.squeeze(0)


def adjacency_matrix(parents: torch.Tensor, directed: bool = True, direction_up: bool = False, loop_root: bool = False):
    """
    Generates the adjacency matrix for a parents array.

    Currently this works only for single arrays (so no batching).

    :param parents: Array of parents: `tree.parents`
    :param directed: Should the adjacency matrix be directed?
    :param direction_up: Per default the root points at its children.
    :param loop_root: Should root have a loop.
    :return:
    """
    root_val = parents[0].item()
    parents[0] = 0

    a = torch.arange(parents.size(-1), device=parents.device)
    adj = parents.new_zeros(parents.size(-1), parents.size(-1), dtype=torch.bool)
    if directed and direction_up:
        adj[a, parents] = True
    elif directed and not direction_up:
        adj[parents, a] = True
    else:
        adj[a, parents] = True
        adj[parents, a] = True

    parents[0] = root_val

    if not loop_root:
        adj[0, 0] = 0

    return adj


def incidences_to_nodes(node_incidences: torch.Tensor, pad_idx=-1):
    """
    This function turns the binary incidence_matrix into a long tensor
     containing the actual position of the node in the sequence.

    Example:
    >>> incidences = node_incidence_matrix(
    >>>    descendants=torch.tensor(
    >>>        [[6, 2, 4, 2, 2],
    >>>         [1, 5, 2, 3, 2],
    >>>         [1, 5, 2, 3, 2]]
    >>>    ),
    >>>    pad_idx=1
    >>> )

    >>> incidences
    >>> tensor([[[ True, False, False, False, False],
    >>>          [ True,  True, False, False, False],
    >>>          [ True, False,  True, False, False],
    >>>          [ True, False,  True,  True, False],
    >>>          [ True, False,  True, False,  True]],
    >>>         [[False, False, False, False, False],
    >>>          [False,  True, False, False, False],
    >>>          [False,  True,  True, False, False],
    >>>          [False,  True, False,  True, False],
    >>>          [False,  True, False,  True,  True]],
    >>>         [[False, False, False, False, False],
    >>>          [False,  True, False, False, False],
    >>>          [False,  True,  True, False, False],
    >>>          [False,  True, False,  True, False],
    >>>          [False,  True, False,  True,  True]]])

    OUTPUT:
    >>> incidences_to_nodes(incidences, pad_idx=-1)
    >>> tensor([[[ 0, -1, -1, -1, -1],
    >>>          [ 0,  1, -1, -1, -1],
    >>>          [ 0, -1,  2, -1, -1],
    >>>          [ 0, -1,  2,  3, -1],
    >>>          [ 0, -1,  2, -1,  4]],
    >>>         [[-1, -1, -1, -1, -1],
    >>>          [-1,  1, -1, -1, -1],
    >>>          [-1,  1,  2, -1, -1],
    >>>          [-1,  1, -1,  3, -1],
    >>>          [-1,  1, -1,  3,  4]],
    >>>         [[-1, -1, -1, -1, -1],
    >>>          [-1,  1, -1, -1, -1],
    >>>          [-1,  1,  2, -1, -1],
    >>>          [-1,  1, -1,  3, -1],
    >>>          [-1,  1, -1,  3,  4]]])


    :param node_incidences: Bool Tensor as returend from node_incidence_matrix
    :param pad_idx:
    :return:
    :rtype:
    """
    assert node_incidences.ndimension() <= 3, "Wrong dimensions on descendants tensor"
    batched = node_incidences.ndimension() == 3
    B = node_incidences.size(0) if batched else 1
    S = node_incidences.size(-1)

    if not batched:
        node_incidences = node_incidences[None,:,:]

    nodes = torch.arange(S).to(node_incidences.device, dtype=torch.long) # [S]
    nodes = nodes[None, None, :]  # 1, 1, S

    # node_incidences += (pad_idx + 1)
    nodes = nodes.expand_as(node_incidences).clone()  # [B, S, S]
    nodes[~node_incidences] = pad_idx

    if not batched:
        nodes = nodes.squeeze(0)

    return nodes


def levels(node_incidences: torch.Tensor) -> torch.LongTensor:
    """ Computes the level of each node based on a binary node incidence matrix.

    Root has level 0.

    For this tree:
        0
        ├──  1
        └──  2
            ├──  3
            └──  4
                └──  5

    With the following node incidence matrix:

    tensor([[1, 0, 0, 0, 0, 0],
            [1, 1, 0, 0, 0, 0],
            [1, 0, 1, 0, 0, 0],
            [1, 0, 1, 1, 0, 0],
            [1, 0, 1, 0, 1, 0],
            [1, 0, 1, 0, 1, 1]], dtype=torch.uint8)

    Will return:

    tensor([0, 1, 1, 2, 2, 3])
    """
    return node_incidences.sum(dim=-1) - 1

# def adjacency_matrix_batched(parents: torch.Tensor, directed: bool = False, direction_up: bool = False, pad_mask=None):
#     raise NotImplementedError("Wip")
#     batched = parents.ndim == 2
#     if not batched:
#         parents = parents[None, :]
#
#     B, S = parents.shape
#
#     if batched and pad_mask is None:
#         pad_mask = parents < -1
#
#     parents[:, 0], root_vals = 0, parents[:, 0]  # only if right padded
#
#     print("root_vals", root_vals)
#     print("->", parents)
#
#     a = torch.arange(S, device=parents.device)
#     b = torch.arange(B, device=parents.device)
#     adj = parents.new_zeros(B, S, S, dtype=torch.bool)
#
#     if directed and direction_up:
#         adj[a, parents] = True
#     elif directed and not direction_up:
#         adj[parents] = True
#     else:
#         adj[a,parents] = True
#         adj[parents, a] = True
#
#     parents[:, 0] = root_vals
#
#     return adj

def ancestral_matrix(node_incidences: torch.Tensor) -> torch.LongTensor:
    """
    Computes the level of the least common ancestor between any node pair
    (or in other words the length of the common prefix between two nodes).

    For this tree:
        0
        ├──  1
        └──  2
            ├──  3
            └──  4
                └──  5

    With the following node incidence matrix:

    tensor([[1, 0, 0, 0, 0, 0],
            [1, 1, 0, 0, 0, 0],
            [1, 0, 1, 0, 0, 0],
            [1, 0, 1, 1, 0, 0],
            [1, 0, 1, 0, 1, 0],
            [1, 0, 1, 0, 1, 1]], dtype=torch.uint8)

    This method will return:

    tensor([[0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0],
            [0, 0, 1, 1, 1, 1],
            [0, 0, 1, 2, 1, 1],
            [0, 0, 1, 1, 2, 2],
            [0, 0, 1, 1, 2, 3]])

    Pytorch supports matrix multiplications on cuda tensors only with float tensors,
     thus this method will return a FloatTensor when the node incidence matrix is located
     on cuda. Otherwise it tries to keep the original dtype when possible, otherwise
     returns a LongTensor.

    :param node_incidences: [B, S, S] or [S, S] matrix of node incidences
    :return:
    """
    batched = node_incidences.ndimension() == 3
    node_incidences = to_matmul_compatibility(node_incidences)

    if batched:
        res = torch.bmm(
            node_incidences,
            node_incidences.transpose(1, 2)
        ) - 1
    else:
        res = (node_incidences @ node_incidences.t()) - 1

    return res.long()


def movements(node_incidences: torch.Tensor, pad_idx: int = -1, pad_mask: Optional[torch.Tensor] = None) -> torch.LongTensor:
    """
    Computes movements between nodes.

    For this tree:
        0
        ├──  1
        └──  2
            ├──  3
            └──  4
                └──  5

    With the following node incidence matrix:
    tensor([[1, 0, 0, 0, 0, 0],
            [1, 1, 0, 0, 0, 0],
            [1, 0, 1, 0, 0, 0],
            [1, 0, 1, 1, 0, 0],
            [1, 0, 1, 0, 1, 0],
            [1, 0, 1, 0, 1, 1]], dtype=torch.uint8)

    Will return the following movements between nodes.
    tensor([[0, 0, 0, 0, 0, 0],
            [1, 0, 1, 1, 1, 1],
            [1, 1, 0, 0, 0, 0],
            [2, 2, 1, 0, 1, 1],
            [2, 2, 1, 1, 0, 0],
            [3, 3, 2, 2, 1, 0]])

    You can read it the following:
    To walk from node i to node j one needs to make
     1. res[i, j] upward movements
     2. res[j, i] downward movements

    Some examples:
     - from 0 to 4 -> res[0,4]=0; res[4,0]=2 (0 steps up, 2 steps down)
     - from 1 to 5 -> res[1,5]=1; res[5,1]=3 (1 step  up, 3 steps down)

    Will always return a LongTensor.
    """
    batched = node_incidences.ndimension() == 3
    B = node_incidences.size(0) if batched else 1
    S = node_incidences.size(1)
    start = pad_idx + 1

    # on cuda we need to have a FloatTensor to do matmuls
    # so lets do that one time in advance
    if node_incidences.is_cuda:
        node_incidences = node_incidences.float()

    node_levels = levels(node_incidences) - 1
    length_of_common_prefix = ancestral_matrix(node_incidences)

    if batched:
        moves = node_levels[:, None, :] - length_of_common_prefix
    else:
        moves = node_levels - length_of_common_prefix

    moves = moves.transpose(-2, -1)  #
    moves = moves.long()
    moves += start

    if pad_mask is not None:
        moves = apply_pad_mask_to_higher_dim_tensor(moves, pad_idx, pad_mask)

    return moves


def distances(
        node_incidences: torch.Tensor, pad_idx: int = -1, pad_mask: Optional[torch.Tensor] = None
) -> torch.LongTensor:
    """
    Calculates distances between every node and every other node in an undirected tree.

    Distances will begin at (pad_idx + 1), to support padded batches.
    """
    # batch size x sequence length
    batched = node_incidences.ndimension() == 3
    S = node_incidences.size(1)
    start = pad_idx + 1

    if not batched:
        node_incidences = node_incidences[None, :, :]  # add batch dim

    inverted_node_incidences = ~node_incidences

    # matmul on cuda works only with float tensors
    node_incidences = to_matmul_compatibility(node_incidences)
    inverted_node_incidences = to_matmul_compatibility(inverted_node_incidences)

    # we want to compute the amount of unequal elements between two pathes
    # (XOR between the pathes and afterwards sum the positive bits).
    distances = torch.bmm(node_incidences, inverted_node_incidences.transpose(1, 2))  # [B, S, S]
    distances += torch.bmm(inverted_node_incidences, node_incidences.transpose(1, 2))  # [B, S, S]

    distances = distances.long()
    distances += start

    if not batched:
        distances = distances.squeeze(0)  # remove batch dim

    if pad_mask is not None:
        distances = apply_pad_mask_to_higher_dim_tensor(
            distances,
            pad_idx,
            pad_mask
        )

    return distances


def least_common_ancestors(node_incidences: torch.Tensor) -> torch.LongTensor:
    """
    Computes least common ancestors for a 1D tensor of num descendants
    :param node_incidences:  shape [S, S]. Does not work when batched
    :return:
    """
    assert node_incidences.ndimension() == 2

    # FIXME/TODO  make it work with padded batches, currently only works with 1D
    S = node_incidences.size(-1)
    nodes = incidences_to_nodes(
        node_incidences,
        pad_idx=(-10 * S)  # Needs to be something small, otherwise wont work
    )

    # this works by comparing a path of a node with another path.
    # consider these two pathes:
    # [    0,     1,     2,     3, -1000, -1000, -1000, -1000],
    # [    0,     1,     2, -1000,     4, -1000, -1000, -1000]
    # if we add those together the LCA (2) will have the highest value.
    # [    0,     2,     4,  -997,  -996, -2000, -2000, -2000]
    #                    ^
    ancestors = torch.argmax(
        nodes[:, None, :] + nodes,  # [S, S, S]  expensive!
        dim=-1
    )  # [S, S]

    return ancestors


def parents_from_descendants(descendants: Sequence[int]) -> torch.Tensor:
    """ not very performant, but it works."""

    stack_idx = [0]
    stack_open_descendants = [descendants[0]]

    parents = [-1]

    for original_idx, num_descendants in enumerate(descendants[1:], start=1):
        parents.append(stack_idx[-1])

        stack_idx.append(original_idx)
        stack_open_descendants.append(num_descendants + 1)

        stack_open_descendants = [d - 1 for d in stack_open_descendants if (d - 1) > 0]
        stack_idx = stack_idx[:len(stack_open_descendants)]

    return descendants.new_tensor(
        parents
    ) if isinstance(descendants, torch.Tensor) else torch.tensor(parents, dtype=torch.long)


def descendants_from_parents(parents: Sequence[int]) -> torch.Tensor:
    descendants = parents.new_zeros(
        parents.size(-1)
    ) if isinstance(parents, torch.Tensor) else torch.zeros(len(parents), dtype=torch.long)

    active = torch.full_like(descendants, fill_value=False, dtype=torch.bool)  # bool tensor with all false

    for node_idx, parent_idx in enumerate(parents):
        active[(parent_idx + 1):] = False  # deactivate closed branch
        descendants[active] += 1  # increment descendants on all active nodes
        active[node_idx] = True  # set current node as active

    return descendants


def descendants_from_node_incidences(node_incidences: torch.Tensor):
    """ Computes the descendants array from a (batched) node incidence matrix.

    If the matrix is batched, the returned descendants array will have -1 on padded indices.
    """
    return node_incidences.sum(-2) - 1


def delete_subtree(tree: TensorTree, node_idx: Union[int, torch.Tensor], replacement: Optional[Any] = None) -> TensorTree:
    """
    Returns a new tree with branch at node_idx deleted or replaced with a single node without children.

    Does the following, given this tree and node_idx=2, replacement_token=99:

    0 MethodDeclaration
    ├── 1 parameters
    │   ├── 2 FormalParameter
    │   │   ├── 3 type
    │   │   │   └── 4 ReferenceType
    │   │   │       └── 5 name
    │   │   │           └── 6 Bitmap
    │   │   └── 7 name
    │   │       └── 8 bmp
    │   └── 9 FormalParameter
    │       ├── 10 type
    │       │   └── 11 ReferenceType
    │       │       └── 12 name
    │       │           └── 13 File
    │       └── 14 name
    └── 15 body

    Will return tensors for the following tree:

    0 MethodDeclaration (0)
    ├── 1 parameters (1)
    │   ├── 2 <MASK> (2)
    │   └── 9 FormalParameter (3)
    │       ├── 10 type (4)
    │       │   └── 11 ReferenceType (5)
    │       │       └── 12 name (6)
    │       │           └── 13 File (7)
    │       └── 14 name (8)
    └── 15 body (9)

    The original tensors have been:
    tokens:        [  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15]
    parents:       [ -1,  0,  1,  2,  3,  4,  5,  2,  7,  1,  9, 10, 11, 12,  9,  0]
    #descendants:  [ 14, 13,  6,  3,  2,  1,  0,  1,  0,  5,  3,  2,  1,  0,  0,  0]
    #children:     [  2,  2,  2,  1,  1,  1,  0,  1,  0,  2,  1,  1,  1,  0,  0,  0]

    This method will return the following tensors:
    words:        [  0,  1, 99,  9, 10, 11, 12, 13, 14, 15]
    parents:      [ -1,  0,  1,  1,  3,  4,  5,  6,  3,  0]
    #descendants: [  9,  7,  0,  5,  3,  2,  1,  0,  0,  0]
    #children:    [  2,  2,  0,  2,  1,  1,  1,  0,  0,  0]
    """

    if replacement is not None:
        assert isinstance(replacement, (torch.Tensor, int, type(tree.node_data[node_idx]))), "Replacement token needs to be tensor type"

    delete_all_children = replacement is None

    # which nodes to keep
    num_removed_nodes = tree.descendants[node_idx].item()

    if delete_all_children:
        num_removed_nodes += 1

    indices_to_keep = torch.cat(
        (
            torch.arange(node_idx if delete_all_children else node_idx + 1),  # keep node_idx
            torch.arange(node_idx + num_removed_nodes if delete_all_children else node_idx + 1 + num_removed_nodes, len(tree))
        )
    )

    # select indices to keep from tokens
    if isinstance(tree.node_data, (torch.Tensor, np.ndarray)):
        node_data = tree.node_data[indices_to_keep]
    else:
        # handle lists
        node_data = [tree.node_data[index.item()] for index in indices_to_keep]

    # and from parent and descendant tensors
    parents = tree.parents[indices_to_keep]
    descendants = tree.descendants[indices_to_keep]

    # explicitly set masked token
    if not delete_all_children:
        descendants[node_idx] = 0
        node_data[node_idx] = replacement

    # Adjust parents after new mask_pos
    parents[node_idx + 1:][parents[node_idx + 1:] > node_idx + 1] -= num_removed_nodes

    # go through each parent of node at mask_pos and adjust descendants
    for ancestor in tree.iter_ancestors(node_idx):
        descendants[ancestor] -= num_removed_nodes

    return tensortree.tree(node_data=node_data, parents=parents, descendants=descendants)


def delete_children(
        tree: TensorTree, node_idx: int, replacement_token: Any = 999,
) -> TensorTree:
    """
    Parents is expected to be a tensor, which can be used to index the tensors (remove any padding or
    incrementations due to padding before !). Roots parent idx should be -1.

    Does the following, given this tree and node_idx=2, mask_token=99:

    0 MethodDeclaration
    ├── 1 parameters
    │   ├── 2 FormalParameter
    │   │   ├── 3 type
    │   │   │   └── 4 ReferenceType
    │   │   │       └── 5 name
    │   │   │           └── 6 Bitmap
    │   │   └── 7 name
    │   │       └── 8 bmp
    │   └── 9 FormalParameter
    │       ├── 10 type
    │       │   └── 11 ReferenceType
    │       │       └── 12 name
    │       │           └── 13 File
    │       └── 14 name
    └── 15 body

    Will return tensors for the following tree:

    0 MethodDeclaration (0)
    ├── 1 parameters (1)
    │   ├── 2 FormalParameter (2)
    │   │   └── 3 <MASK> (3)
    │   └── 9 FormalParameter (4)
    │       ├── 10 type (5)
    │       │   └── 11 ReferenceType (6)
    │       │       └── 12 name (7)
    │       │           └── 13 File (8)
    │       └── 14 name (9)
    └── 15 body (10)

    The original tensors have been:
    tokens:        [  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15]
    parents:       [ -1,  0,  1,  2,  3,  4,  5,  2,  7,  1,  9, 10, 11, 12,  9,  0]
    #descendants:  [ 14, 13,  6,  3,  2,  1,  0,  1,  0,  5,  3,  2,  1,  0,  0,  0]
    #children:     [  2,  2,  2,  1,  1,  1,  0,  1,  0,  2,  1,  1,  1,  0,  0,  0]

    This method will return the following tensors:
    words:        [ 0, 1, 2, 99, 9, 10, 11, 12, 13, 14, 15]
    parents:      [-1, 0, 1,  2, 1,  4,  5,  6,  7,  4,  0]
    #descendants: [10, 8, 1,  0, 5,  3,  2,  1,  0,  0,  0]
    #children:    [ 2, 2, 1,  0, 2,  1,  1,  1,  0,  0,  0]

    """
    if replacement_token is not None:
        assert isinstance(replacement_token,
                          (torch.Tensor, int, type(tree.node_data[node_idx]))), "Replacement token needs to be tensor type"

    pos_mask = node_idx + 1
    num_descendants_of_masked_node = tree.descendants[node_idx]
    num_removed_nodes = num_descendants_of_masked_node - 1

    assert num_descendants_of_masked_node > 0, "Node to mask children must have at least 1 child!"

    indices_to_keep = torch.cat(
        [
            torch.arange(node_idx + 2),  # keep node_idx and 1 child (which we will mask)
            torch.arange(node_idx + num_descendants_of_masked_node + 1, tree.descendants.size(0))
        ]
    )
    parents = tree.parents[indices_to_keep]
    descendants = tree.descendants[indices_to_keep]

    if is_tensor_type(tree.node_data):
        node_data = tree.node_data[indices_to_keep]
    else:
        # handle other sequence types, such as list of strings
        node_data = [tree.node_data[index.item()] for index in indices_to_keep]

    node_data[pos_mask] = replacement_token
    descendants[node_idx] = 1
    descendants[pos_mask] = 0

    # Adjust parents after new mask_pos
    parents[node_idx + 1:][parents[node_idx + 1:] > node_idx + 1] -= num_removed_nodes

    # go through each parent of node at mask_pos and adjust descendants
    for ancestor in tree.iter_ancestors(node_idx):
        descendants[ancestor] -= num_removed_nodes

    return tensortree.tree(node_data=node_data, parents=parents, descendants=descendants)


def swap(tree: TensorTree, node_1: Union[int, torch.Tensor], node_2: Union[int, torch.Tensor]):
    """
    Does the following given this tree and 1, 8 for the swap positions:

    before swap:

    0 MethodDeclaration
    ├── 1 BasicType
    │   └── 2 int
    ├── 3 name
    │   └── 4 roundToDimensions
    ├── 5 FormalParameter
    │   ├── 6 BasicType
    │   │   └── 7 int
    │   └── 8 value
    ├── 9 FormalParameter
    │   ├── 10 BasicType
    │   │   └── 24 int
    │   └── 11 nearestValue
    ...

    after the swap:

    0 MethodDeclaration
    ├── 1 value
    ├── 2 name
    │   └── 3 roundToDimensions
    ├── 4 FormalParameter
    │   ├── 5 BasicType
    │   │   └── 6 int
    │   └── 7 BasicType
    │       └── 8 int
    ├── 9 FormalParameter
    │   ├── 10 BasicType
    │   │   └── 11 int
    │   └── 12 nearestValue
    ...
    The original tensors have been:
        tokens:        [ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ...]
        parents:       [-1,  0,  1,  0,  3,  0,  5,  6,  5,  0,  9, 10,  9, ...]
        #descendants:  [27,  1,  0,  1,  0,  3,  1,  0,  0,  3,  1,  0,  0, ...]
        #children:     [ 5,  1,  0,  1,  0,  2,  1,  0,  0,  2,  1,  0,  0, ...]

    This method will return the following tensors:
        words:        [ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ...]
        parents:      [-1,  0,  1,  0,  3,  0,  5,  6,  5,  0,  9, 10,  9, ...]
        #descendants: [27,  1,  0,  1,  0,  3,  1,  0,  0,  3,  1,  0,  0, ...]
        #children:    [ 5,  1,  0,  1,  0,  2,  1,  0,  0,  2,  1,  0,  0, ...]


    Here a sketch to understand the meaning of the variables defined further down:

                        tree_start_diff
                     <---------------------------->
    +----------------+--------+-------------------+------------------------+---------+
    |trees be4 tree 1| tree 1 |t. between swap t. |        tree 2          |t. after | <--- Tensor before swap
    +----------------+--------+-------------------+------------------------+---------+
                     ^        ^                   ^                        ^
                     |        |                   |                        |
                     +        +                   +                        +
                  node_1    node_1_end          node_2               node_2_end
                     +  node_2_swp_end       node_1_swp_start              +
                     |        +                   +                        |
                     |        |                   |                        |
                     v        +---------------v   +---------------v        v
    +----------------+------------------------+-------------------+--------+---------+
    |trees be4 tree 1|        tree 2          |t. between swap t. | tree 1 |t. after | <--- Tensor after swap
    +----------------+------------------------+-------------------+--------+---------+
                                                  <--------------->
                                                  tree_length_diff

    """

    if node_1 == node_2:  # nothing to do
        return tree.detach()
    elif node_1 > node_2:
        node_1, node_2 = node_2, node_1

    # + 1's are for easier slicing
    size = len(tree)
    tree_length_diff = tree.get_number_of_descendants(node_1) - tree.get_number_of_descendants(node_2)
    tree_start_diff = node_2 - node_1

    node_1_end = tree.next_node_not_in_branch(node_1)
    node_2_end = tree.next_node_not_in_branch(node_2)

    print("node_1_end, node_2_end:", node_1_end, node_2_end)
    # node_1_end = node_1 + tree.get_number_of_descendants(node_1)
    # node_2_end = node_2 + tree.get_number_of_descendants(node_2) + 1

    node_1_swp_start = node_2 - tree_length_diff
    node_2_swp_end = node_1 + tree.get_number_of_descendants(node_2) + 1

    # the start of the first swapped tree and the end of
    # the second swapped tree do not change
    # node_1_end += 1  # easier slicing

    # Check if the second tree is a sub-tree of the first one
    if ((node_1_end - 1) if node_1_end is not None else len(tree) + 1) >= node_2:
        raise ValueError(f"Node {node_2} is a subtree of node {node_1}")

    def swap_node_data() -> torch.Tensor:
        original_node_data = tree.node_data
        swapped_node_data = original_node_data.clone()  # create a copy

        swapped_node_data[node_1_swp_start:node_2_end] = original_node_data[node_1:node_1_end]  # copy subtree 1
        swapped_node_data[node_2_swp_end:node_1_swp_start] = original_node_data[node_1_end:node_2]  # move nodes between trees
        swapped_node_data[node_1:node_2_swp_end] = original_node_data[node_2:node_2_end]  # copy subtree 2
        return swapped_node_data

    def swap_parents() -> torch.Tensor:
        original_parents = tree.parents
        swapped_parents = original_parents.clone()

        swapped_parents[node_1_swp_start + 1:node_2_end] = original_parents[node_1 + 1:node_1_end]  # copy subtree 1
        swapped_parents[node_1_swp_start + 1:node_2_end] += tree_start_diff - tree_length_diff  # adjust

        # since the trees should swap we have to skip the first parents
        beginning_2 = node_2 + 1
        beginning_2_swp = node_1_swp_start + 1
        # all Nodes which parents are between the swap-trees
        mask = original_parents[node_1_end:beginning_2] > node_1

        # parent points to node be4 swap (no adjustments) or after swap (needs to be adjusted)
        swapped_parents[node_2_swp_end:beginning_2_swp][mask] = \
            original_parents[node_1_end:beginning_2][mask] - tree_length_diff
        swapped_parents[node_2_swp_end:beginning_2_swp][~mask] = \
            original_parents[node_1_end:beginning_2][~mask]

        # copy tree 2
        swapped_parents[node_1 + 1:node_2_swp_end] = \
            original_parents[node_2 + 1:node_2_end] - tree_start_diff

        # all Nodes after the swap which parents are between the swap-trees
        mask1 = (node_1_end - 1) < original_parents[node_2_end:size]
        mask2 = original_parents[node_2_end:size] < node_2
        swapped_parents[node_2_end:size][mask1 & mask2] -= tree_length_diff

        return swapped_parents

    def swap_descendants(swapped_parents) -> torch.Tensor:
        original_descendants = tree.descendants
        swapped_descendants = original_descendants.clone()

        # copy tree 1
        swapped_descendants[node_1_swp_start:node_2_end] = original_descendants[node_1:node_1_end]

        # Nodes between swap-trees
        swapped_descendants[node_2_swp_end:node_1_swp_start] = original_descendants[node_1_end:node_2]

        # copy tree 2
        swapped_descendants[node_1:node_2_swp_end] = original_descendants[node_2:node_2_end]

        # update the descendant count of the ancestors (note the ancestors did not change, so we can iter the
        # original tree)
        for ancestor in tree.iter_ancestors(node_1):
            swapped_descendants[ancestor] -= tree_length_diff

        for ancestor in tree.iter_ancestors(node_2):
            swapped_descendants[ancestor] += tree_length_diff

        return swapped_descendants

    node_data = swap_node_data()
    parents = swap_parents()
    descendants = swap_descendants(parents)

    return tensortree.tree(
        parents=parents,
        descendants=descendants,
        node_data=node_data
    )


def insert_child(
        tree: TensorTree, parent_idx: int, child_node_data: Union[Any, TensorTree],
        right_sibling_idx: Optional[int] = None
) -> TensorTree:
    assert isinstance(
        child_node_data,
        (torch.Tensor, int, type(tree.node_data[parent_idx]), TensorTree)
    ), f"Replacement token needs to be tensor type"

    if right_sibling_idx is None:
        # get node after this branch
        idx_to_insert = tree.next_node_not_in_branch(parent_idx)
        if idx_to_insert is None:
            idx_to_insert = len(tree)
    else:
        idx_to_insert = right_sibling_idx
        if parent_idx != tree.get_parent(right_sibling_idx):
            raise IndexError(f"node at right_sibling_idx needs to have parent_idx as parent and not {tree.get_parent(right_sibling_idx)}")

    if isinstance(child_node_data, TensorTree):
        parents_to_insert = child_node_data.parents.clone()
        parents_to_insert[0] = parent_idx
        parents_to_insert[1:] += idx_to_insert

        descendants_to_insert = child_node_data.descendants
        node_data_to_insert = child_node_data.node_data

    else:
        parents_to_insert = tree.parents.new_tensor([parent_idx])
        descendants_to_insert = tree.descendants.new_tensor([0])

        if isinstance(tree.node_data, torch.Tensor):
            node_data_to_insert = tree.node_data.new_tensor([child_node_data])
        else:
            # handle lists
            node_data_to_insert = [child_node_data]

    # node data
    if isinstance(tree.node_data, torch.Tensor):
        node_data = torch.cat([
            tree.node_data[:idx_to_insert],
            node_data_to_insert,
            tree.node_data[idx_to_insert:],
        ])
    elif isinstance(tree.node_data, list):
        node_data = tree.node_data[:idx_to_insert] + node_data_to_insert + tree.node_data[idx_to_insert:]
    else:
        raise ValueError("Unknown type of node_data.")

    num_nodes_added = parents_to_insert.shape[0]

    # parents
    parents_after_insert = tree.parents[idx_to_insert:].clone()
    parents_after_insert[parents_after_insert >= idx_to_insert] += num_nodes_added

    parents = torch.cat([
        tree.parents[:idx_to_insert],
        parents_to_insert,
        parents_after_insert,
    ])

    descendants = torch.cat([
        tree.descendants[:idx_to_insert],
        descendants_to_insert,
        tree.descendants[idx_to_insert:],
    ])

    # go through each parent of node at mask_pos and adjust descendants
    descendants[parent_idx] += num_nodes_added
    for ancestor in tree.iter_ancestors(parent_idx):
        descendants[ancestor] += num_nodes_added

    return tensortree.tree(node_data=node_data, parents=parents, descendants=descendants)





from dataclasses import dataclass
from typing import Sequence, Any, Union, List, Optional, Tuple, Generator, Callable, Literal

import numpy as np
import torch

import tensortree
from tensortree.render import Style, ContRoundStyle, format_tree
from tensortree.utils import to_torch, validate_index
from functools import lru_cache


# Define a type alias for the content of the node sequence
LabelType = Any
TensorType = Union[Sequence[int], np.ndarray, torch.Tensor]


@dataclass(frozen=True)
class TreeStorage:
    """
    Stores a tree with data.
    """

    # either parents or descendants may be None.
    # other sequence types will be converted to tensors.
    parents: TensorType = None
    descendants: Union[torch.Tensor, Sequence[int]] = None

    # some operations (swapping) only work when node_data is a torch.Tensor
    node_data: Union[torch.Tensor, Sequence[LabelType]] = None

    format: Literal['torch', 'numpy'] = 'torch'

    def __post_init__(self):
        if self.parents is None and self.descendants is None:
            raise ValueError("Either parents or descendants must be passed")

        if self.parents is None:  # compute parents from descendants
            descendants: torch.Tensor = to_torch(self.descendants).long()
            parents = tensortree.parents_from_descendants(self.descendants)

        elif self.descendants is None:  # compute descendants from parents
            parents: torch.Tensor = to_torch(self.parents).long()
            descendants = tensortree.descendants_from_parents(parents)

        else:  # convert everything to a tensor
            parents: torch.Tensor = to_torch(self.parents).long()
            descendants: torch.Tensor = to_torch(self.descendants).long()

        # node_data may be nothing, in that case simply enumerate the nodes
        if self.node_data is None:
            node_data = torch.arange(len(descendants)).to(descendants)
            # FIXME make node_data optional
        else:
            # node_data is a sequence of strings (tensor incompatible)
            try:
                node_data: torch.Tensor = to_torch(self.node_data).long()
            except (ValueError, TypeError, RuntimeError):
                node_data: List[LabelType] = list(self.node_data)

        if descendants.numel() != len(node_data) != parents.numel():
            raise ValueError(f"All arrays need to be of same length and not ({descendants.numel()}, {len(node_data)}, {parents.numel()}).")

        object.__setattr__(self, 'parents', parents)
        object.__setattr__(self, 'descendants', descendants)
        object.__setattr__(self, 'node_data', node_data)


def tree(
    parents: Optional[TensorType] = None,
    node_data: Optional[Sequence[LabelType]] = None,
    descendants: Optional[TensorType] = None,
):
    """ Constructor to build a tree. """

    return TensorTree.from_array(parents=parents, descendants=descendants, node_data=node_data)


class TensorTree:

    @classmethod
    def from_array(
            cls,
            parents: Optional[TensorType] = None, descendants: Optional[TensorType] = None,
            node_data: Optional[Sequence[LabelType]] = None
    ):
        """ Obtain a tree from arrays. A tree can be either defined by a parents or a descendants tensor.
        Additional nodes list can be passed if it contains a string, it will be used for rendering.
        """
        return cls(TreeStorage(parents, descendants, node_data))

    @classmethod
    def from_node_incidences(cls, node_incidences: torch.Tensor, node_data: Optional[Sequence[LabelType]] = None):
        descendants = tensortree.descendants_from_node_incidences(node_incidences)
        return cls.from_array(descendants=descendants, node_data=node_data)

    def __init__(self, data: TreeStorage, root_idx: int = 0):
        """
        Initialize with a pointer to tree storage.
        """
        self.data = data
        self.root_idx = root_idx

        if len(self.data.parents) > 0 and (self.data.parents[0] != -1):  # and self.data.parents[0] != 0):
            raise ValueError("Parents array seems to have wrong format.")

        self.__len = self.data.descendants.shape[-1]  # cache this

        # span in original array
        self.end = len(self)

        if self.is_subtree():
            self.__len = self.data.descendants[root_idx] + 1
            self.end = root_idx + len(self)

    def __len__(self):
        """ The number of nodes in this tree. """
        return self.__len

    def __str__(self):
        return self.pformat(max_nodes=10)

    @validate_index(allow_none=True)
    def __getitem__(self, node_idx: Union[int, torch.Tensor, np.ndarray, Tuple[int, ...], None]):
        """ Will returns a view of the node at node_idx"""
        if node_idx is None and node_idx != self.root_idx:
            return TensorTree(self.data)  # take me back to the very root

        node_idx = self._to_global_idx(node_idx)

        if node_idx == self.root_idx:
            return self

        return TensorTree(self.data, node_idx)

    def detach(self):
        """ Returns a new tree rooted at self.root_idx """
        from copy import deepcopy
        return self.from_array(
            parents=self.parents.clone(),
            descendants=self.descendants.clone(),
            node_data=deepcopy(self.node_data)
        )

    def is_subtree(self) -> bool:
        """
        Is this tree instance part of a bigger tree?

        You can retrieve the original tree using
        >>> root = subtree[None]
        """

        return self.root_idx > 0

    def is_descendant_of(self, node_ancestor_idx: int, node_descendant_idx: int) -> bool:
        """
        Is one node a descendant of another node?

        :param node_ancestor_idx: The node closer to the root
        :param node_descendant_idx: The descendant
        :return:
        """
        return node_ancestor_idx < node_descendant_idx < (
                node_ancestor_idx + self.get_number_of_descendants(node_ancestor_idx)
        )

    # helpers for individual nodes
    @validate_index
    def get_node_data(self, node_idx: Union[int, torch.Tensor]) -> Any:
        """
        Returns the data of a node.

        :param node_idx: The index of the node.
        """
        node_idx = self._to_global_idx(node_idx)
        return self.data.node_data[node_idx]

    @validate_index
    def get_number_of_descendants(self, node_idx: Union[int, torch.Tensor]) -> torch.Tensor:
        """
        Returns the amount of descendants of a node.

        :param node_idx:
        :return:
        """
        node_idx = self._to_global_idx(node_idx)
        return self.data.descendants[node_idx]

    @validate_index
    def get_parent(self, node_idx: Union[int, torch.Tensor]) -> Optional[int]:
        """ Returns the parent idx for this node or None if node_idx is root.

        :param node_idx: Node_idx refers to the index of the node in the original tree.
        :return:
        """
        node_idx = self._to_global_idx(node_idx)

        # return None for root
        if node_idx == self.root_idx:
            return

        return self.data.parents[node_idx]

    @property
    def descendants(self) -> torch.Tensor:
        """
        Returns the relevant subset of descendants for this subtree. This will return
         a slice of the data. Changing this object will change the storage and may
         invalidate the tree. This is not checked.
        """
        if self.root_idx == 0:
            return self.data.descendants  # return full array
        else:
            return self.data.descendants[self.root_idx: self.end]  # return slice

    @property
    def node_data(self) -> Sequence[LabelType]:
        """
        Returns the relevant subset of node_data for this tree. This will return
         a slice of the data. Changing this object will change the storage and may
         invalidate the tree. This is not checked.

        """
        if self.root_idx == 0:
            return self.data.node_data
        else:
            return self.data.node_data[self.root_idx: self.end]

    @property
    def parents(self) -> torch.Tensor:
        """
        Returns the relevant subset of parents for this subtree. This returns the underlying
         storage only if it is root. Otherwise will create a new array for this subtree.

        """
        if self.root_idx == 0:
            return self.data.parents

        # compute parents
        if getattr(self, '__subtree_parents', None) is None:
            parents = self.data.parents[self.root_idx: self.end] - self.root_idx
            parents[0] = -1
            self.__subtree_parents = parents  # cache parents for this subtree

        return self.__subtree_parents

    @lru_cache(maxsize=1)
    def node_incidence_matrix(self) -> torch.BoolTensor:
        """ Returns the node incidence matrix for this tree"""
        return tensortree.node_incidence_matrix(self.descendants)

    @lru_cache(maxsize=1)
    def adjacency_matrix(
            self, directed: bool = True, direction_up: bool = False, loop_root: bool = False) -> torch.BoolTensor:
        """
         Returns the adjacency matrix for this tree.

        :param directed: Should the adjacency matrix be directed?
        :param direction_up: Per default the root points at its children.
        :param loop_root: Should root have a loop.
        :return:
        """
        return tensortree.adjacency_matrix(
            parents=self.parents,
            directed=directed, direction_up=direction_up,
            loop_root=loop_root
        )

    def ancestral_matrix(self) -> torch.LongTensor:
        """
        Computes the level of the least common ancestor between any node pair
        (or in other words the length of the common prefix between two nodes).

        For this tree:
            0
            ├──  1
            └──  2
                ├──  3
                └──  4
                    └──  5

        With the following node incidence matrix:

        tensor([[1, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0],
                [1, 0, 1, 1, 0, 0],
                [1, 0, 1, 0, 1, 0],
                [1, 0, 1, 0, 1, 1]], dtype=torch.uint8)

        This method will return:

        tensor([[0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 1],
                [0, 0, 1, 2, 1, 1],
                [0, 0, 1, 1, 2, 2],
                [0, 0, 1, 1, 2, 3]])

        Pytorch supports matrix multiplications on cuda tensors only with float tensors,
         thus this method will return a FloatTensor when the node incidence matrix is located
         on cuda. Otherwise it tries to keep the original dtype when possible, otherwise
         returns a LongTensor.

        :return:
        """

        return tensortree.ancestral_matrix(self.node_incidence_matrix())

    def movements(self) -> torch.LongTensor:
        """
        Computes movements between nodes.

        For this tree:
            0
            ├──  1
            └──  2
                ├──  3
                └──  4
                    └──  5

        With the following node incidence matrix:
        tensor([[1, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0],
                [1, 0, 1, 1, 0, 0],
                [1, 0, 1, 0, 1, 0],
                [1, 0, 1, 0, 1, 1]], dtype=torch.uint8)

        Will return the following movements between nodes.
        tensor([[0, 0, 0, 0, 0, 0],
                [1, 0, 1, 1, 1, 1],
                [1, 1, 0, 0, 0, 0],
                [2, 2, 1, 0, 1, 1],
                [2, 2, 1, 1, 0, 0],
                [3, 3, 2, 2, 1, 0]])

        You can read it the following:
        To walk from node i to node j one needs to make
         1. res[i, j] upward movements
         2. res[j, i] downward movements

        Some examples:
         - from 0 to 4 -> res[0,4]=0; res[4,0]=2 (0 steps up, 2 steps down)
         - from 1 to 5 -> res[1,5]=1; res[5,1]=3 (1 step  up, 3 steps down)

        Will always return a LongTensor.
        """
        return tensortree.movements(self.node_incidence_matrix())

    def distances(self) -> torch.LongTensor:
        """
        Computes distances between nodes when the tree is seen as undirected.

        For this tree:
            0
            ├──  1
            └──  2
                ├──  3
                └──  4
                    └──  5

        With the following node incidence matrix:
        tensor([[1, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0],
                [1, 0, 1, 1, 0, 0],
                [1, 0, 1, 0, 1, 0],
                [1, 0, 1, 0, 1, 1]], dtype=torch.uint8)

        Will return the following distances between nodes.
        tensor([[0, 1, 1, 2, 2, 3],
                [1, 0, 2, 3, 3, 4],
                [1, 2, 0, 1, 1, 2],
                [2, 3, 1, 0, 2, 3],
                [2, 3, 1, 2, 0, 1],
                [3, 4, 2, 3, 1, 0]])

        Will always return a LongTensor.
        """
        return tensortree.distances(self.node_incidence_matrix())

    def least_common_ancestors(self):
        """
        Return a least common ancestor matrix for this tree.

        :return:
        """
        return tensortree.least_common_ancestors(self.node_incidence_matrix())

    def levels(self) -> torch.LongTensor:
        """
        Computes the level of each node (i.e. the number of edges from a node to the root).
        Root has level 0.

        For this tree:
            0
            ├──  1
            └──  2
                ├──  3
                └──  4
                    └──  5

        With the following node incidence matrix:

        tensor([[1, 0, 0, 0, 0, 0],
                [1, 1, 0, 0, 0, 0],
                [1, 0, 1, 0, 0, 0],
                [1, 0, 1, 1, 0, 0],
                [1, 0, 1, 0, 1, 0],
                [1, 0, 1, 0, 1, 1]], dtype=torch.uint8)

        Will return:

        tensor([0, 1, 1, 2, 2, 3])
        """
        return tensortree.levels(self.node_incidence_matrix())

    @validate_index
    def next_node_not_in_branch(self, node_idx: Union[int, torch.Tensor]) -> Optional[torch.Tensor]:
        """
        Return the next node, that is not part of the branch of node at node_idx.

        Can be a sibling or any other node that follows after this branch.

        :param node_idx: Can be either an integer or a single value tensor
        :return:
        """
        next_node_idx = (node_idx + self.get_number_of_descendants(node_idx) + 1)
        return next_node_idx if next_node_idx < len(self) else None

    @validate_index
    def step_out(self, node_idx: Union[int, torch.Tensor]) -> Optional[torch.Tensor]:
        """
        Return the next node, that is not part of the subtree of node_idx's parent.

        Can be a sibling to nodes parent or any other node that follows.

        :param node_idx: Can be either an integer or a single value tensor
        :return:
        """
        nodes_parent = self.get_parent(node_idx)
        if nodes_parent is None:
            return

        return self.next_node_not_in_branch(nodes_parent)

    def node_idx_for_tree_position(self, tree_position: Tuple[int, ...]) -> Union[int, torch.Tensor]:
        if not isinstance(tree_position, tuple):
            raise ValueError("tree_position must be tuple")

        def get_nth_child(node_idx, n):
            for i, child_idx in enumerate(self.iter_children(node_idx)):
                if i == n:
                    return child_idx
            raise IndexError

        current_node_idx = 0
        for child_number in tree_position:
            current_node_idx = get_nth_child(current_node_idx, child_number)

        return current_node_idx

    def heights(self):
        """ Depth of each subtree"""
        heights = torch.zeros_like(self.descendants)

        for i, layer_mask in enumerate(tensortree.mask_layer(self.node_incidence_matrix())):
            heights[layer_mask] = i

        return heights

    # subtree mask
    @validate_index
    def subtree_mask(self, node_idx: Union[int, torch.Tensor]) -> torch.BoolTensor:
        """ returns a mask which selects the relevant nodes of a subtree from the array """

        start = node_idx
        end = self.next_node_not_in_branch(node_idx)

        mask = torch.zeros_like(self.descendants, dtype=torch.bool)
        mask[start:end] = True
        return mask

    # leaves
    @validate_index
    def is_leaf(self, node_idx: Union[int, torch.Tensor]) -> bool:
        return self.get_number_of_descendants(node_idx) == 0

    def leaves_mask(self) -> torch.BoolTensor:
        """ Returns a boolean mask for all leaf nodes in this tree """
        return self.descendants == 0

    def leaf_indices(self) -> torch.Tensor:
        return self.leaves_mask().nonzero(as_tuple=False).squeeze(-1)

    # children
    @validate_index
    def iter_children(self, node_idx: Union[int, torch.Tensor]) -> Generator[torch.Tensor, None, None]:
        """
        Iters over the children of a node with a specific index in a tree.

        :param node_idx: Node to iterate over the children
        """
        branch_end = self.next_node_not_in_branch(node_idx)  # end of subtree of node_idx

        if branch_end is None:
            branch_end = len(self)

        # next child is at the next position in the descendants array
        next_child = node_idx + 1

        # are we still in the subtree
        while next_child is not None and next_child < branch_end:
            yield next_child
            next_child = self.next_node_not_in_branch(next_child)

    @validate_index
    def children_mask(self, node_idx: Union[int, torch.Tensor]) -> torch.BoolTensor:
        """
        Returns a mask over the child indices of a node.

        :param node_idx: Node to get the children
        """
        return self.adjacency_matrix()[node_idx]

    @validate_index
    def children(self, node_idx: Union[int, torch.Tensor]) -> torch.Tensor:
        """
        Returns children indices of a node at a specific index in a tree.

        :param node_idx: Node to get the children
        """
        return self.children_mask(node_idx).nonzero(as_tuple=False).squeeze(-1)

    # siblings
    @validate_index
    def iter_siblings(self, node_idx: Union[int, torch.Tensor], include_left_siblings: bool = True, include_right_siblings: bool = True) -> Generator[torch.Tensor, None, None]:
        """ Node indices with the same parent. The node at node_idx is excluded.

        set include_left_siblings to False to only output siblings which are to the right of node at node_idx.
        set include_right_siblings to False  to only output siblings which are to the left of node at node_idx.
        """
        if not (include_left_siblings or include_right_siblings):
            raise ValueError("Left or right must be specified.")

        parent_idx = self.get_parent(node_idx)

        # root
        if parent_idx is None:
            return

        reached_node = False
        for node in self.iter_children(parent_idx):
            if node == node_idx:
                reached_node = True
                continue

            if include_left_siblings and not reached_node:
                yield node

            if include_right_siblings and reached_node:
                yield node

    @validate_index
    def right_sibling(self, node_idx: Union[int, torch.Tensor]) -> Optional[torch.Tensor]:
        next_node = self.next_node_not_in_branch(node_idx)
        if next_node is not None:
            if self.get_parent(next_node) == self.get_parent(node_idx):
                return next_node

    @validate_index
    def siblings_mask(self, node_idx: Union[int, torch.Tensor], include_left_siblings: bool = True, include_right_siblings: bool = True) -> torch.BoolTensor:
        """ Node indices with the same parent. The node at node_idx is excluded.

        set include_left_siblings to False to only output siblings which are to the right of node at node_idx.
        set right to False  to only output siblings which are to the left of node at node_idx.
        """
        if not (include_left_siblings or include_right_siblings):
            raise ValueError("Left or right must be specified.")

        parent_idx = self.get_parent(node_idx)

        # root
        if parent_idx is None:
            return self.data.descendants.new_zeros(len(self)).bool()

        all_siblings = self.children_mask(parent_idx)

        assert all_siblings[node_idx], "this should be true"
        # exlude node at node_idx
        all_siblings[node_idx] = False

        # exclude everything before node_idx
        if not include_left_siblings:
            all_siblings[:node_idx] = False

        # exclude everything after node_idx
        if not include_right_siblings:
            all_siblings[node_idx + 1:] = False

        return all_siblings

    @validate_index
    def siblings(self, node_idx: Union[int, torch.Tensor], check_left: bool = True, check_right: bool = True) -> torch.Tensor:
        return self.siblings_mask(node_idx, check_left, check_right).nonzero(as_tuple=False).squeeze(-1)

    @validate_index
    def has_sibling(self, node_idx: Union[int, torch.Tensor], check_left: bool = True, check_right: bool = True) -> bool:
        """
        Is there a sibling to this node?

        :param node_idx:
        :param check_left: look left for siblings (set to False to only check for right siblings)
        :param check_right: look right for siblings (set to False to only check for left siblings)
        :return:
        """
        node_parent = self.get_parent(node_idx)
        if node_parent is None:
            return False

        result = False
        if check_right:
            next_node = self.next_node_not_in_branch(node_idx)
            if next_node is not None:
                next_nodes_parent = self.get_parent(next_node)
                has_right_sibling = (next_nodes_parent == node_parent)
                result = result or has_right_sibling
        if check_left and not result:
            # otherwise something should be between parent and node
            has_left_sibling = (node_parent + 1) != node_idx
            result = result or has_left_sibling

        return result

    # ancestors
    @validate_index
    def iter_ancestors(self, node_idx: Union[int, torch.Tensor]) -> Generator[torch.Tensor, None, None]:
        node_parent = self.get_parent(node_idx)
        while node_parent is not None:
            yield node_parent
            node_parent = self.get_parent(node_parent)

    # pretty printing
    def pformat(
            self, max_nodes: Optional[int] = None, node_renderer: Callable[[Any], str] = str,
            style: Union[Style] = ContRoundStyle,
    ) -> str:
        """
        Pretty prints a tree up to `max_nodes`. Define a node_renderer for custom node types (e.g. Dictionaries).
        :param max_nodes: Render up to this amount of nodes.
        :param node_renderer: A function that outputs a string.
        :param style: Style the tree.
        :return:
        """

        return format_tree(tree=self, max_nodes=max_nodes, node_renderer=node_renderer, style=style)

    def pprint(
            self, max_nodes: Optional[int] = None, node_renderer: Callable[[Any], str] = str,
            style: Union[Style] = ContRoundStyle,
    ):
        """ See pformat for description of arguments."""
        print("TensorTree():\n", self.pformat(max_nodes, node_renderer, style))

    @validate_index
    def _to_global_idx(self, node_idx: Union[int, torch.Tensor]) -> Union[int, torch.Tensor]:
        """ Transfers a node_idx inside a subtree view to a global node_idx """
        return node_idx + self.root_idx

    # functions below return modify and return a new tree
    @validate_index
    def delete_node(self, node_idx: int, replacement_token: Optional[Any] = None):
        """
            Returns a new tree with branch at node_idx deleted or replaced with a single node without children.

            Does the following, given this tree and node_idx=2, replacement_token=99:

            0 MethodDeclaration
            ├── 1 parameters
            │   ├── 2 FormalParameter
            │   │   ├── 3 type
            │   │   │   └── 4 ReferenceType
            │   │   │       └── 5 name
            │   │   │           └── 6 Bitmap
            │   │   └── 7 name
            │   │       └── 8 bmp
            │   └── 9 FormalParameter
            │       ├── 10 type
            │       │   └── 11 ReferenceType
            │       │       └── 12 name
            │       │           └── 13 File
            │       └── 14 name
            └── 15 body

            Will return tensors for the following tree:

            0 MethodDeclaration (0)
            ├── 1 parameters (1)
            │   ├── 2 <MASK> (2)
            │   └── 9 FormalParameter (3)
            │       ├── 10 type (4)
            │       │   └── 11 ReferenceType (5)
            │       │       └── 12 name (6)
            │       │           └── 13 File (7)
            │       └── 14 name (8)
            └── 15 body (9)

            For node_idx=2, replacement_token=None:

            0 MethodDeclaration (0)
            ├── 1 parameters (1)
            │   └── 9 FormalParameter (3)
            │       ├── 10 type (4)
            │       │   └── 11 ReferenceType (5)
            │       │       └── 12 name (6)
            │       │           └── 13 File (7)
            │       └── 14 name (8)
            └── 15 body (9)


            The original tensors have been:
            nodes:        [  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15]
            parents:       [ -1,  0,  1,  2,  3,  4,  5,  2,  7,  1,  9, 10, 11, 12,  9,  0]
            #descendants:  [ 14, 13,  6,  3,  2,  1,  0,  1,  0,  5,  3,  2,  1,  0,  0,  0]
            #children:     [  2,  2,  2,  1,  1,  1,  0,  1,  0,  2,  1,  1,  1,  0,  0,  0]

            This method will return the following tensors:
            nodes:        [  0,  1, 99,  9, 10, 11, 12, 13, 14, 15]
            parents:      [ -1,  0,  1,  1,  3,  4,  5,  6,  3,  0]
            #descendants: [  9,  7,  0,  5,  3,  2,  1,  0,  0,  0]
            #children:    [  2,  2,  0,  2,  1,  1,  1,  0,  0,  0]
            """
        return tensortree.delete_subtree(self, node_idx, replacement_token)

    @validate_index
    def delete_children(self, node_idx: int, replacement_token: Optional[Any] = None):
        return tensortree.delete_children(self, node_idx, replacement_token)

    @validate_index
    def swap(self, node_idx: int, other_node_idx: int):
        return tensortree.swap(self, node_idx, other_node_idx)

    @validate_index
    def insert_child(self, parent_idx: int, node_data: Any, right_sibling_idx: Optional[int] = None):
        """ adds a node (or a TensorTree) as a child of node at parent_idx, so that it is the left sibling of
         node at right_sibling_idx. If right_sibling_idx is None then it will be appended as the last child."""

        return tensortree.insert_child(self, parent_idx, node_data, right_sibling_idx)






from setuptools import setup, find_namespace_packages

setup(
    name='tensortree',
    packages=["tensortree"],
    version='0.2',
    python_requires='>=3.7',
    description='Represent trees with tensors and perform fast calculations and modifications.',
    author='Johannes Villmow',
    author_email='johannes.villmow@hs-rm.de',
    url='https://github.com/villmow/tensortree',
    download_url='https://github.com/villmow/tensortree/archive/refs/tags/v0.2.tar.gz',
    license='MIT',
    install_requires=[
        "numpy",
        "torch",
    ],
    keywords=['tree', 'tensor', 'pytorch', 'pytorch tree', 'tensor tree'],  # Keywords that define your package best
    classifiers=[
        'Development Status :: 4 - Beta',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
)

